/*
 * HTFileNode.java
 * www.bouthier.net
 *
 * The MIT License :
 * -----------------
 * Copyright (c) 2001 Christophe Bouthier
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package hypertree.demo;

import hypertree.HTNode;
import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.awt.Color;
import java.util.Date;


/**
 * The HTFileNode implements an example of HTNode encapsulating a File.
 * <P>
 * The color legend is :
 * <UL>
 *   <IL> white  for files less than a hour old
 *   <IL> green  for files less than a day old
 *   <IL> yellow for files less than a week old
 *   <IL> orange for files less than a month old
 *   <IL> red    for files less than a year old
 *   <IL> blue   for files more than a year old
 * </UL>
 *
 * @author Christophe Bouthier [bouthier@loria.fr]
 * @version 1.0
 */
public class HTFileNode
    implements HTNode {

    private File      file     = null; // the File encapsulated
    private Hashtable children = null; // the children of this node

	public File getFile(){return file;};

  /* --- Constructor --- */

    /**
     * Constructor.
     *
     * @param file    the File encapsulated in this node
     */
    public HTFileNode(File file) {
        this.file = file;
        children = new Hashtable();

        if (! isLeaf()) {
            String[] tabFichiers = file.list();
            for (int i = 0; i < tabFichiers.length; i++) {
                File fichier = new File(file.getPath() +
                                        File.separator + tabFichiers[i]);
                HTFileNode child = new HTFileNode(fichier);
                addChild(child);
            }

        }
    }


  /* --- Tree management --- */

    /**
     * Add child to the node.
     * 
     * @param child    the HTFileNode to add as a child
     */
    protected void addChild(HTFileNode child) {
        children.put(child.getName(), child);
    }


  /* --- HTNode --- */

    /**
     * Returns the children of this node in an Enumeration.
     * If this node is a file, return a empty Enumeration.
     * Else, return an Enumeration full with HTFileNode.
     *
     * @return    an Enumeration containing childs of this node
     */
    public Enumeration children() {
        return children.elements();
    }

    /**
     * Returns true if this node is not a directory.
     *
     * @return    <CODE>false</CODE> if this node is a directory;
     *            <CODE>true</CODE> otherwise
     */
    public boolean isLeaf() {
        return (! file.isDirectory());
    }
    
    /**
     * Returns the name of the file.
     *
     * @return    the name of the file
     */
    public String getName() {
        return file.getName();
    }

    /**
     * Returns the color of the file.
     *
     * @return    the color of the file
     */
    public Color getColor() {
        long time = file.lastModified();
        long diff = (new Date()).getTime() - time;

        if (diff <= 3600000L) {             // less than an hour
            return Color.white;
        } else if (diff <= 86400000L) {     // less than a day
            return Color.green;
        } else if (diff <= 604800000L) {    // less than a week
            return Color.yellow;
        } else if (diff <= 2592000000L) {   // less than a month
            return Color.orange;
        } else if (diff <= 31536000000L) {  // less than a year
            return Color.red;
        } else {                            // more than a year
            return Color.blue;
        }
    }
}

