/*
 * HTNewickParser.java
 * www.bouthier.net
 *
 * The MIT License :
 * -----------------
 * Copyright (c) 2003 Christophe Bouthier
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package hypertree.demo;

import java.io.*;
import java.util.Vector;


/**
 * The HTNewickParser class implements a Newick tree format parser.
 *
 * @author Christophe Bouthier [bouthier@loria.fr]
 * @version 1.0
 */
public class HTNewickParser {

    /* --- States --- */
    private static final int STATE_ERROR           = -1;
    private static final int STATE_INITIAL         = 0;
    private static final int STATE_COMMENT_I       = 1;
    private static final int STATE_CREATION_F      = 2;
    private static final int STATE_CREATION_L      = 3;
    private static final int STATE_START_NAME      = 4;
    private static final int STATE_COMMENT_N       = 5;
    private static final int STATE_NAME_U          = 6;
    private static final int STATE_NAME_Q          = 7;
    private static final int STATE_Q_STOP          = 8;
    private static final int STATE_ADD_Q           = 9;
    private static final int STATE_SPACE_EATER_N   = 10;
    private static final int STATE_COMMENT_N_L     = 11;
    private static final int STATE_START_L         = 12;
    private static final int STATE_COMMENT_L       = 13;
    private static final int STATE_L_N             = 14;
    private static final int STATE_L_F             = 15;
    private static final int STATE_SPACE_EATER_L   = 16;
    private static final int STATE_COMMENT_END_L   = 17;
    private static final int STATE_COMMIT          = 18;
    private static final int STATE_COMMIT_CLOSE_F  = 19;
    private static final int STATE_COMMIT_END      = 20;
    private static final int STATE_ADD_L_SIGNE     = 21;

    
    /* --- Character types --- */
    private static final int TYPE_OTHER            = 0;
    private static final int TYPE_WHITESPACE       = 1;
    private static final int TYPE_CONTROL          = 2;
    private static final int TYPE_DIGIT            = 3;
    private static final int TYPE_LEFT_PARENTHES   = 4;
    private static final int TYPE_RIGHT_PARENTHES  = 5;
    private static final int TYPE_LEFT_SQ_BRACKET  = 6;
    private static final int TYPE_RIGHT_SQ_BRACKET = 7;
    private static final int TYPE_COMMA            = 8;
    private static final int TYPE_DOT              = 9;
    private static final int TYPE_SEMI_COLON       = 10;
    private static final int TYPE_COLON            = 11;
    private static final int TYPE_SINGLE_QUOTE     = 12;
    private static final int TYPE_PLUS_SIGNE       = 13;
    private static final int TYPE_MINUS_SIGNE      = 14;
    

    /* --- Transitions table ---- */
    private static final int[][] transitions = {
      //  0   1   2   3   4   5   6   7   8   9  10  11  12  13  14
        {-1,  0, -1, -1,  2, -1,  1, -1, -1, -1, -1, -1, -1, -1, -1}, //  0
        { 1,  1,  1,  1,  1,  1,  1,  0,  1,  1,  1,  1,  1,  1,  1}, //  1
        { 3,  3,  3,  3,  2,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3}, //  2
        { 6,  4, -1,  6,  2, 19,  5, -1, 18,  6, 20, 12,  7,  6,  6}, //  3
        { 6,  4, -1,  6,  2, 19,  5, -1, 18,  6, 20, 12,  7,  6,  6}, //  4
        { 5,  5,  5,  5,  5,  5,  5,  4,  5,  5,  5,  5,  5,  5,  5}, //  5
        { 6, 10, -1,  6, -1, 19, 11, -1, 18,  6, 20, 12, -1,  6,  6}, //  6
        { 7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  8,  7,  7}, //  7
        {-1, 10, -1, -1, -1, 19, 11, -1, 18, -1, 20, 12,  9, -1, -1}, //  8
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, //  9
        {-1, 10, -1, -1, -1, 19, 11, -1, 18, -1, 20, 12, -1, -1, -1}, // 10
        {11, 11, 11, 11, 11, 11, 11, 10, 11, 11, 11, 11, 11, 11, 11}, // 11
        {-1, 12, -1, 14, -1, -1, 13, -1, -1, 15, -1, -1, -1, 21, 21}, // 12
        {13, 13, 13, 13, 13, 13, 13, 12, 13, 13, 13, 13, 13, 13, 13}, // 13
        {-1, 16, -1, 14, -1, 19, 17, -1, 18, 15, 20, -1, -1, -1, -1}, // 14
        {-1, 16, -1, 15, -1, 19, 17, -1, 18, -1, 20, -1, -1, -1, -1}, // 15
        {-1, 16, -1, -1, -1, 19, 17, -1, 18, -1, 20, -1, -1, -1, -1}, // 16
        {17, 17, 17, 17, 17, 17, 17, 16, 17, 17, 17, 17, 17, 17, 17}, // 17
        { 3,  3,  3,  3,  2,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3}, // 18
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, // 19
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, // 20
        {-1, -1, -1, 14, -1, -1, -1, -1, -1, 15, -1, -1, -1, -1, -1}  // 21
    };


    private static final StringBuffer nameDef  = new StringBuffer("default");
    private static final StringBuffer lNameDef = new StringBuffer("default");

    
    private Vector       oldFathers = new Vector();  // old father nodes
    private HTNewickNode father     = null;          // current father node
    private HTNewickNode current    = null;          // current node
    
    private int          state      = STATE_INITIAL; // parser state
    private char         c          = '0';           // character read
    private int          level      = 0;             // deepness level

    private StringBuffer name       = nameDef;       // node name
    private StringBuffer lName      = lNameDef;      // node length
    

  /* --- Constructor --- */
    
    /**
     * Constructor.
     */
    public HTNewickParser() {
    }


  /* --- Parsing --- */

    /**
     * Parse the given file and returns the root of the builded tree.
     *
     * @param file    the file to parse
     * @return        the root of the builded tree
     */
    public HTNewickNode buildTree(File file) {
        try {
            BufferedReader bs = new BufferedReader(new FileReader(file));
            while(state != STATE_COMMIT_END) {
                int i = bs.read();
                if (i == -1) {
                    System.err.println("Error : end of file too early");
                    System.exit(1);
                }
                c = (char) i;
                state = doTransition(state, getCharType(c));
                execute();
            }
        } catch (IOException e) {
            System.err.println("Error with file : " + e);
        }
        return father;
    }

    /**
     * Executing the transaction and returning the new state.
     *
     * @param state    the current state
     * @param type     the type of character received
     * @return         the new state
     */
    private int doTransition(int state,
                             int type) {
        return transitions[state][type];
    }

    /**
     * Returns the type of the given char.
     *
     * @param c    the given char
     * @return     its type
     */
    private int getCharType(char c) {
        if (Character.isWhitespace(c)) {
            return TYPE_WHITESPACE;
        }
        if (Character.isISOControl(c)) {
            return TYPE_CONTROL;
        }
        if (Character.isDigit(c)) {
            return TYPE_DIGIT;
        }
        if (c == '(') {
            return TYPE_LEFT_PARENTHES;
        }
        if (c == ')') {
            return TYPE_RIGHT_PARENTHES;
        }
        if (c == '[') {
            return TYPE_LEFT_SQ_BRACKET;
        }
        if (c == ']') {
            return TYPE_RIGHT_SQ_BRACKET;
        }
        if (c == ',') {
            return TYPE_COMMA;
        }
        if (c == '.') {
            return TYPE_DOT;
        }
        if (c == ';') {
            return TYPE_SEMI_COLON;
        }
        if (c == ':') {
            return TYPE_COLON;
        }
        if (c == '\'') {
            return TYPE_SINGLE_QUOTE;
        }
        if (c == '+') {
            return TYPE_PLUS_SIGNE;
        }
        if (c == '-') {
            return TYPE_MINUS_SIGNE;
        }
        return TYPE_OTHER;
    }

    /**
     * Do the thing to do for this state.
     */
    private void execute() {
        switch(state) {
        case STATE_ERROR:
            System.out.println("Error state in parsing.");
            System.exit(0);
            break;
        case STATE_INITIAL:
            break;
        case STATE_COMMENT_I:
            break;
        case STATE_CREATION_F:
            oldFathers.add(level, father);
            father = new HTNewickNode(false);
            if (level != 0) {
                ((HTNewickNode) oldFathers.elementAt(level)).addChild(father);
            }
            level++;
            break;
        case STATE_CREATION_L:
            current = new HTNewickNode();
            father.addChild(current);
            name = nameDef;
            lName = lNameDef;
            state = doTransition(state, getCharType(c));
            execute();
            break;
        case STATE_START_NAME:
            name = nameDef;
            lName = lNameDef;
            break;
        case STATE_COMMENT_N:
            break;
        case STATE_NAME_U:
            if (name == nameDef) {
                name = new StringBuffer();
            }
            if (c == '_') {
                name.append(' ');
            } else {
                name.append(c);
            }
            break;
        case STATE_NAME_Q:
            if (name == nameDef) {
                name = new StringBuffer();
            }
            if (c != '\'') {
                name.append(c);
            }
            break;
        case STATE_Q_STOP:
            break;
        case STATE_ADD_Q:
            name.append('\'');
            state = STATE_NAME_Q;
            break;
        case STATE_SPACE_EATER_N:
            break;
        case STATE_COMMENT_N_L:
            break;
        case STATE_START_L:
            break;
        case STATE_COMMENT_L:
            break;
        case STATE_L_N:
            if (lName == lNameDef) {
                lName = new StringBuffer();
            }
            lName.append(c);
            break;
        case STATE_L_F:
            if (lName == lNameDef) {
                lName = new StringBuffer();
            }
            lName.append(c);
            break;
        case STATE_SPACE_EATER_L:
            break;
        case STATE_COMMENT_END_L:
            break;
        case STATE_COMMIT:
            commit();
            break;
        case STATE_COMMIT_CLOSE_F:
            commit();
            level--;
            current = father;
            if (level != 0) {
                father = (HTNewickNode) oldFathers.elementAt(level);
            }
            state = STATE_START_NAME;
            execute();
            break;
        case STATE_COMMIT_END:
            commit();
            if (level != 0) {
                System.err.println("Error ';' before level 0");
                System.exit(0);
            }
            break;
        case STATE_ADD_L_SIGNE:
            if (lName == lNameDef) {
                lName = new StringBuffer();
            }
            lName.append(c);
            break;
        default:
            break;
        }
    }

    /**
     * Commit the node changes.
     */
    private void commit() {
        if (name != nameDef) {
            current.setName(name.toString());
        }

        if (lName != lNameDef) {
            try {
                current.setLength(Double.parseDouble(lName.toString()));
            } catch (NumberFormatException e) {
                System.err.println("Error in a length parsing: " + e);
            }
        }
    }

}

